// Simple driver that demonstrates dynamically loading and unloading

#include "ntddk.h"
#include "cancel.h"


//
// Global Data
//



//
// function prototypes
//

NTSTATUS
CancelOpen(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    );

NTSTATUS
CancelClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    );

VOID
CancelUnload(
    IN PDRIVER_OBJECT DriverObject
    );

NTSTATUS
CancelReadWrite(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    );

NTSTATUS
CancelIoctl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    );

VOID CancelTimerDpc(
    IN PKDPC Dpc,
    IN PVOID DeferredContext,
    IN PVOID SystemArgument1,
    IN PVOID SystemArgument2
    );

VOID  CancelCancel ( 
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp 
    );

//
// Data structure definitions, macros, inline functions
//

//
// don't like asserts that dont work, so we use our own.
//
__inline void _osrBugCheck(char * file, int line)
{
    KdPrint(("Assertion failed in file %s at line %d\n", file, line));
    KeBugCheckEx( 0x22 , line, 0, 0, 0);
}

#define OsrBugCheck() _osrBugCheck(__FILE__, __LINE__)

#define OsrAssertAlways(x) if (!(x)) { OsrBugCheck(); }

#if DBG

#define OsrAssert(x) OsrAssertAlways(x)

#else

#define OsrAssert(x)

#endif 
//
// The whole point of this driver is to park read/write irps on a queue
// and keep them there until the issuing thread terminates. Just to make things
// interesting, every now and then a timer goes off and pops the top irp off
// the queue and completes it.
//
// To represent this activity a device extension is defined that contains a queue
// a lock and the structures required to support a timer dpc routine.
//
typedef struct {

    LIST_ENTRY     irpList;             // our private IRP queue
    ULONG          queueLength;         // how big is it?
    KSPIN_LOCK     lock;                // lock covers this structure;
    KTIMER         timer;               // timer for completing IRPs
    KDPC           dpc;                 // timer dpc object
    PDEVICE_OBJECT device;              // backpointer to the device object
    ULONG          state;               // operational state of the device
    ULONG          speed;               // canonical representation of DPC speed
    LARGE_INTEGER  CancelDpcInterval;   // how fast we run the DPC
    BOOLEAN        takeFromFront;       // complete io from head or tail of queue?
    ULONG          tagfield;            // structure tag

} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

#define DEVICE_EXTENSION_TAG 0x73A00101

//
// validate that the devExtension is non-NULL and that the
// structure tag field has the correct value
//
__inline BOOLEAN validExtension(PDEVICE_EXTENSION devExtension)
{
    if (devExtension && 
        (devExtension->tagfield == DEVICE_EXTENSION_TAG))
    {
        return TRUE;
    }

    return FALSE;
}

//
// produce a device extension pointer from a device object,
// along the way validate that it appears to be OUR device extension
// 
// stronger validation can be achieved by putting a unique tag field
// in the structure.
//
__inline PDEVICE_EXTENSION getDeviceExtension(PDEVICE_OBJECT deviceObject)
{
    PDEVICE_EXTENSION deviceExtension;

    OsrAssert(deviceObject);

    deviceExtension = deviceObject->DeviceExtension;

    OsrAssert(validExtension(deviceExtension) && 
                (deviceExtension->device == deviceObject));

    return deviceExtension;
}

//
// set the initial state of a device extension object.
//
__inline void initDeviceExtension(PDEVICE_EXTENSION devExtension, PDEVICE_OBJECT deviceObject)
{
    InitializeListHead(&devExtension->irpList);
    devExtension->queueLength = 0;

    KeInitializeSpinLock(&devExtension->lock);
    KeInitializeTimer(&devExtension->timer);
    KeInitializeDpc(&devExtension->dpc, CancelTimerDpc, devExtension);

    devExtension->device = deviceObject;
    devExtension->state = CANCEL_ON  | COMPLETE_ON | COMPLETE_CANCELED | SET_CANCEL | EARLY_RELEASE;
    devExtension->speed = MED_SLOW;
    devExtension->CancelDpcInterval.QuadPart = MED_SLOW_SPEED;
    devExtension->takeFromFront = FALSE;
    devExtension->tagfield = DEVICE_EXTENSION_TAG;
}

__inline void setState(ULONG flags, ULONG speed, PDEVICE_EXTENSION devExtension)
{
    KIRQL irql;

    //
    // get the spinlock, then change the state
    //
    KdPrint(("setState acquiring spinlock\n"));
    KeAcquireSpinLock(&devExtension->lock, &irql);

    devExtension->speed = speed;
    
    switch (speed)
    {
    default:
        KdPrint( ( "speed %x? setting to slow\n", speed ));
        speed = SLOW;
        //
        // FALL THROUGH!
    case SLOW:
        devExtension->CancelDpcInterval.QuadPart = SLOW_SPEED;
        break;

    case MED_SLOW:
        devExtension->CancelDpcInterval.QuadPart = MED_SLOW_SPEED;
        break;

    case MED_FAST:
        devExtension->CancelDpcInterval.QuadPart = MED_FAST_SPEED;
        break;

    case FAST:
        devExtension->CancelDpcInterval.QuadPart = FAST_SPEED;
        break;
    }    
    
    devExtension->state = flags;

    KdPrint(("setState releasing spinlock\n"));
    KeReleaseSpinLock(&devExtension->lock, irql);

    KdPrint(("New Flags: %x Completion speed is now %d\n", flags, devExtension->CancelDpcInterval.LowPart));
}

__inline ULONG getQueueLength(PDEVICE_EXTENSION deviceExtension)
{
    return deviceExtension->queueLength;
}



NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{

    PDEVICE_OBJECT deviceObject = NULL;
    NTSTATUS status;
    UNICODE_STRING uniNtNameString;
    UNICODE_STRING uniWin32NameString;
    PDEVICE_EXTENSION deviceExtension = NULL;

    KdPrint( ("Cancel: Entered the cancel driver!\n") );

    //
    // Create counted string version of our device name.
    //

    RtlInitUnicodeString( &uniNtNameString, NT_DEVICE_NAME );

    //
    // Create the device object
    //

    status = IoCreateDevice(
                 DriverObject,
                 sizeof(DEVICE_EXTENSION), // our private data
                 &uniNtNameString,
                 CANCEL_DEVICE_TYPE,
                 0,                     // No standard device characteristics
                 FALSE,                 // This isn't an exclusive device
                 &deviceObject
                 );

    if ( NT_SUCCESS(status) )
    {
        //
        // init the device extension
        //
        deviceExtension = deviceObject->DeviceExtension;
        OsrAssert(deviceExtension);
  
        initDeviceExtension(deviceExtension, deviceObject);

        //
        // Create dispatch points for create/open, close, unload.
        //

        DriverObject->MajorFunction[IRP_MJ_CREATE] = CancelOpen;
        DriverObject->MajorFunction[IRP_MJ_CLOSE] = CancelClose;
        DriverObject->MajorFunction[IRP_MJ_READ] = CancelReadWrite;
        DriverObject->MajorFunction[IRP_MJ_WRITE] = CancelReadWrite;
        DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = CancelIoctl;
        DriverObject->DriverUnload = CancelUnload;

        //
        // make this a direct IO device
        //
        deviceObject->Flags |= DO_DIRECT_IO;

        KdPrint( ("Cancel: just about ready!\n") );

        //
        // Create counted string version of our Win32 device name.
        //
    
        RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME );
    
        //
        // Create a link from our device name to a name in the Win32 namespace.
        //
        
        status = IoCreateSymbolicLink( &uniWin32NameString, &uniNtNameString );

        if (!NT_SUCCESS(status))
        {
            KdPrint( ("Cancel: Couldn't create the symbolic link\n") );

            IoDeleteDevice( DriverObject->DeviceObject );
        }
        else
        {
            KdPrint( ("Cancel: All initialized!\n") );
        }
    }
    else
    {
        KdPrint( ("Cancel: Couldn't create the device\n") );
    }
    return status;
}

NTSTATUS
CancelOpen(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{

    //
    // No need to do anything.
    //

    //
    // Fill these in before calling IoCompleteRequest.
    //
    // DON'T get cute and try to use the status field of
    // the irp in the return status.  That IRP IS GONE as
    // soon as you call IoCompleteRequest.
    //

    KdPrint( ("Cancel: Opened!!\n") );

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return STATUS_SUCCESS;
}

NTSTATUS
CancelClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{

    //
    // No need to do anything.
    //

    //
    // Fill these in before calling IoCompleteRequest.
    //
    // DON'T get cute and try to use the status field of
    // the irp in the return status.  That IRP IS GONE as
    // soon as you call IoCompleteRequest.
    //

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    KdPrint( ("Cancel: Closed!!\n") );

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return STATUS_SUCCESS;
}

VOID
CancelUnload(
    IN PDRIVER_OBJECT DriverObject
    )
{
    UNICODE_STRING uniWin32NameString;

    //
    // All *THIS* driver needs to do is to delete the device object and the
    // symbolic link between our device name and the Win32 visible name.
    //
    // Almost every other driver ever witten would need to do a
    // significant amount of work here deallocating stuff.
    //

    KdPrint( ("Cancel: Unloading!!\n") );
    
    //
    // Create counted string version of our Win32 device name.
    //

    RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME );

    //
    // Delete the link from our device name to a name in the Win32 namespace.
    //
    
    IoDeleteSymbolicLink( &uniWin32NameString );

    //
    // Finally delete our device object
    //

    IoDeleteDevice( DriverObject->DeviceObject );
}

NTSTATUS
CancelReadWrite(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    PDEVICE_EXTENSION devExtension = getDeviceExtension(DeviceObject);
    BOOLEAN needNewDpc = FALSE;
    BOOLEAN queued = FALSE;
    KIRQL irql;
    PETHREAD thread = PsGetCurrentThread();
    unsigned char * buffer;
    //
    // always park Irps on the queue, marking them for cancellation
    // note that we acquire our own spinlock around BOTH operations
    //

    //
    // mark the irp pending 
    //
    IoMarkIrpPending(Irp);
    //
    // lets stash the system virtual address for the MDL that describes
    // the user buffer in IRP->Tail.Overlay.DriverContext[0].
    // 
    // This makes sure that we have mapped the user buffer into system space
    // and now we can see what if anything the operating system has done if
    // we fail to cancel the irp.
    //
    buffer = MmGetSystemAddressForMdl(Irp->MdlAddress);
    Irp->Tail.Overlay.DriverContext[0] = buffer;

    KdPrint(("\tCancelReadWrite: Acquiring spinlock\n"));

    KeAcquireSpinLock(&devExtension->lock, &irql);

    KdPrint(("\tSpinlock acquired for cpu %x\n", KeGetCurrentProcessorNumber()));

    if (devExtension->state & SET_CANCEL) {
        //
        // if the user requested us to do nothing, we don't set the cancel routine
        //
        IoSetCancelRoutine(Irp, CancelCancel);
    }
    //
    // see if the list is empty - if so fire up
    // a new dpc
    //
    needNewDpc = IsListEmpty(&devExtension->irpList);

   	InsertTailList(&devExtension->irpList, &Irp->Tail.Overlay.ListEntry);

    devExtension->queueLength++;

    KdPrint(("\tCancelReadWrite: Irp %x buffer %x thread %x position %d\n", Irp, buffer, thread, devExtension->queueLength));
    
    if (needNewDpc) {

        queued = KeSetTimer( &devExtension->timer, devExtension->CancelDpcInterval, &devExtension->dpc);
        if (TRUE == queued) {
            //
            // this shouldn't happen
            //
            KdPrint( ("\ttimer already queued\n") );
        }
    }
    
    KdPrint(("\tCancelReadWrite releasing spinlock for cpu %x\n", KeGetCurrentProcessorNumber()));

    KeReleaseSpinLock(&devExtension->lock, irql);
    //
    // always return status pending -
    // note that we might very well have already completed,
    // or cancelled this IRP.
    //
    return STATUS_PENDING;
}

NTSTATUS
CancelIoctl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    PDEVICE_EXTENSION devExtension = getDeviceExtension(DeviceObject);
    PIO_STACK_LOCATION IoStack = IoGetCurrentIrpStackLocation(Irp);
    ULONG flags;
    NTSTATUS status = STATUS_SUCCESS;
    PCANCEL_STATE cancelState = NULL;
    ULONG infoSize = 0;

    //
    // we only support one IOCTL: 
    // IOCTL_CANCEL_SET_FLAGS sets a DWORD bitmask that alters the runtime 
    // behavior of the driver.
    //
    switch(IoStack->Parameters.DeviceIoControl.IoControlCode)
    {
    default:
        //
        // some odd IOCTL found its way here
        //
        status = STATUS_UNSUCCESSFUL;
        break;

    case IOCTL_CANCEL_SET_FLAGS:
        //
        // make sure that the input buffer is the correct size
        //
        if (IoStack->Parameters.DeviceIoControl.InputBufferLength != sizeof(CANCEL_STATE) )
        {
            status = STATUS_INFO_LENGTH_MISMATCH; // whatever
            break;

        } 

        // get the flags from the system buffer
        OsrAssert(Irp->AssociatedIrp.SystemBuffer);

        cancelState = (PCANCEL_STATE ) Irp->AssociatedIrp.SystemBuffer;
        //
        // see if the flags are valid
        //
        if (cancelState->flags & (~VALID_FLAGS)) {

            status = STATUS_INVALID_PARAMETER;
			KdPrint(("IOCTL_CANCEL_SET_FLAGS invalid flags %x\n", cancelState->flags));
            break;
        }
        //
        // OK lets just do what it says
        //
        setState(cancelState->flags, cancelState->speed, devExtension);
        
        break;

    case IOCTL_CANCEL_GET_FLAGS:
        //
        // make sure the output buffer is the correct size
        //
        if (IoStack->Parameters.DeviceIoControl.OutputBufferLength != sizeof(CANCEL_STATE) )
        {
            status = STATUS_INFO_LENGTH_MISMATCH; // whatever
            break;

        } 

        // get the flags from the system buffer
        OsrAssert(Irp->AssociatedIrp.SystemBuffer);

        cancelState = (PCANCEL_STATE ) Irp->AssociatedIrp.SystemBuffer;
        
        cancelState->flags = devExtension->state;
        cancelState->speed = devExtension->speed;
        cancelState->queueLength = getQueueLength(devExtension);
        infoSize = sizeof(CANCEL_STATE);
        
        break;

    }

    //
    // now complete the IOCTL
    //
    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = infoSize;

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    return status;
}

__inline BOOLEAN validateWriteBuffer(unsigned char * buffer, ULONG length)
{
    ULONG sequence;
    unsigned char data=255;

    if (length < sizeof(ULONG)) {

        KdPrint(("\t\tBad Length %d\n", length));

        return FALSE;
    }
    sequence = *(PULONG)buffer;
    length -= sizeof(ULONG);
    buffer += sizeof(ULONG);

    while (length-- > 0) {
        if (*buffer != ((data+1)%256)) {

            KdPrint(("\t\tExpected %x got %x\n", ((data+1)%256), *buffer));

            return FALSE;
        }
        data = *buffer++;
    }

    KdPrint(("\t\tbuffer number %d OK\n", sequence));
    return TRUE;
}

VOID CancelTimerDpc(
    IN PKDPC Dpc,
    IN PVOID DeferredContext,
    IN PVOID SystemArgument1,
    IN PVOID SystemArgument2
    )
{
    PDEVICE_EXTENSION devExtension = DeferredContext;
    PIRP Irp;
    PLIST_ENTRY element;
    BOOLEAN queued;
    BOOLEAN complete = TRUE;

    KdPrint(("\t\tCancelTimerDpc getting device lock cpu %x\n", KeGetCurrentProcessorNumber()));    
    KeAcquireSpinLockAtDpcLevel(&devExtension->lock);
    KdPrint(("\t\tCancelTimerDpc lock acquired cpu %x\n", KeGetCurrentProcessorNumber()));  

    if (devExtension->state & COMPLETE_ON) {
        //
        // pluck a single Irp off the queue and complete the poor thing.
        //
        if (!IsListEmpty(&devExtension->irpList)) {            
            
            if (devExtension->takeFromFront) {

                element = RemoveHeadList(&devExtension->irpList);
            
            } else {

                element = RemoveTailList(&devExtension->irpList);

            }
            //
            // alternate completion from the front and back of the queue
            //
            devExtension->takeFromFront = !devExtension->takeFromFront;

            OsrAssert( element && (element != &devExtension->irpList)); // nasty bug if it were
            
            OsrAssert(devExtension->queueLength > 0);
            devExtension->queueLength--;
        
            Irp = CONTAINING_RECORD(element, IRP, Tail.Overlay.ListEntry);
            //
            // See if the Irp is already cancelled
            //
            if (Irp->Cancel) {
                //
                // Hmmmm... what to do?
                //
                KdPrint(("\t\tIrp %x already cancelled in dpc routine\n", Irp));
                if (! (devExtension->state & COMPLETE_CANCELED)) {

                    complete = FALSE;
                }
            }
            //
            // now complete it with STATUS_UNSUCCESSFULL
            //
            if (complete) { 
                //
                // before we actually complete this IRP, lets take a look at the
                // data that the user is outputting.
                //
                // Note that we are assumming that Parameters.Read and Parameters.Write
                // have the same memory layout.
                //
                PIO_STACK_LOCATION ioStack = IoGetCurrentIrpStackLocation(Irp);
                ULONG length = ioStack->Parameters.Write.Length;
                unsigned char * buffer = MmGetSystemAddressForMdl(Irp->MdlAddress);

                if (buffer) {

                    //
                    // compare THIS SVA with the one we stashed in the IRP.
                    //
                    if (buffer != Irp->Tail.Overlay.DriverContext[0]) {

                        KdPrint(("\t\tIrp %x buffer %x expected %x\n", Irp, buffer, Irp->Tail.Overlay.DriverContext[0]));
                    }

                    switch (ioStack->MajorFunction) {
                    case IRP_MJ_WRITE:
                        //
                        // read every byte in the buffer
                        // the first ULONG is a sequence number,
                        // every other byte should just monotonically increment to 256,
                        // and then wrap.
                        //
                        if (FALSE == validateWriteBuffer(buffer, length)) {
                            //
                            // observe that we have accessed a system virtual address
                            // that is in some unknown state.
                            //
                            KdPrint(("\t\tIrp %x Buffer %x invalid write data\n", Irp, buffer));

                        }                     
                        break;

                    case IRP_MJ_READ:
                        //
                        // for a READ we write every byte in the buffer, notice that unlike
                        // the WRITE case we are modifying some system virtual address.
                        // If this address is in an unknown state the system can be corrupted.
                        //
                        while (length-- > 0) {

                            *buffer++ = (unsigned char)(length%256);
                        }                     

                        break;

                    default:
                        //
                        // YIKES
                        //
                        break;
                        KdPrint(("\t\tMajorFunction %x unexpected\n", ioStack->MajorFunction));
                    }

                } else {

                    KdPrint(("\t\tIrp %x can't get SystemAddressForMdl %x\n", Irp, Irp->MdlAddress));

                }



                KdPrint(("\t\tCompleting Irp %x on CPU %x queuelength %d\n", Irp, KeGetCurrentProcessorNumber(), devExtension->queueLength));

                //
                // always remove cancel routine 
                // or we bugcheck in free build, assert in checked
                //
                (void) IoSetCancelRoutine(Irp, NULL);
                //
                // indicate we are finished
                //
                Irp->IoStatus.Status = STATUS_SUCCESS;
                //
                // no actual data transfer.
                //
                Irp->IoStatus.Information = 0;
                //
                // complete the request
                //
                IoCompleteRequest(Irp, IO_NO_INCREMENT);
            }
        
        } else {

            OsrAssert(devExtension->queueLength == 0);
        }
    }

    //
    // If there are more irps, reschedule yourself
    //
    if (!IsListEmpty(&devExtension->irpList)) {
        //
        // queue that funky timer white boy
        //
        OsrAssert(devExtension->queueLength > 0);
        queued = KeSetTimer( &devExtension->timer, devExtension->CancelDpcInterval, &devExtension->dpc);

        if (TRUE == queued) {
            //
            // this shouldn't happen
            //
            KdPrint( ("timer already queued\n") );
        }

    } else {

        OsrAssert(devExtension->queueLength == 0);
    }

    KdPrint(("\t\tCancelTimerDpc releasing device lock\n"));
    KeReleaseSpinLockFromDpcLevel(&devExtension->lock);

}

VOID  CancelCancel ( 
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp 
    )
{
    PDEVICE_EXTENSION devExtension = DeviceObject->DeviceExtension;
    PLIST_ENTRY nextEl = NULL;
    PIRP cancelIrp = NULL;
    KIRQL irql;
    KIRQL cancelIrql = Irp->CancelIrql;
    ULONG position = 0;

    OsrAssert(devExtension);

    KdPrint(("\t\t\tCancelCancel Irp %x CancelIrql %x\n", Irp, cancelIrql));
    //
    // if the DONT_CANCEL flag is set then we don't do anything here
    //
    if (!(devExtension->state & CANCEL_ON)) {
        //
        // just bail out
        //
        IoReleaseCancelSpinLock(cancelIrql);
        KdPrint(("\t\t\tCANCEL_ON not set - cancel does nothing\n"));
        return;
    }
    //
    // release the cancel spinlock now?
    //
    if (devExtension->state & EARLY_RELEASE) {
     
        IoReleaseCancelSpinLock(cancelIrql);
    }
    
    //
    // A thread has terminated and we should find a
    // cancelled Irp in our queue and complete it
    // 
    KdPrint(("\t\t\tCancelCancel acquiring device lock cpu %x\n", KeGetCurrentProcessorNumber()));
    KeAcquireSpinLock(&devExtension->lock, &irql);
    KdPrint(("\t\t\tCancelCancel lock acquired cpu %x irql %x\n", KeGetCurrentProcessorNumber(), irql));

    // 
    // search our queue for an Irp to cancel
    //
    for (nextEl = devExtension->irpList.Flink; 
         nextEl != &devExtension->irpList; )  
    {
         OsrAssert(devExtension->queueLength > 0);

         cancelIrp = CONTAINING_RECORD(nextEl, IRP, Tail.Overlay.ListEntry);

         nextEl = nextEl->Flink;

         position++;

         OsrAssert(position <= devExtension->queueLength);

         if (cancelIrp->Cancel) {
             //
             // if we are ONLY cancelling the passed in IRP, test this now
             //
             if (devExtension->state & CANCEL_IRP) {
                 //
                 // dont cancel this irp
                 //
                 if (Irp != cancelIrp) {
                     //
                     // indicate we have still done nothing
                     //
                     cancelIrp = NULL;

                     continue;

                 }
             }      
             //
             // dequeue THIS irp
             //
             RemoveEntryList(&cancelIrp->Tail.Overlay.ListEntry);

             devExtension->queueLength--;
             //
             // and stop right here
             //
             break;
         }
         cancelIrp = NULL;
    }
    
    KdPrint(("\t\t\tCancelCancel releasing device lock cpu %x\n", KeGetCurrentProcessorNumber()));
    KeReleaseSpinLock(&devExtension->lock, irql);
    //
    // release the cancel spinlock before completing the cancelled Irp.
    // but only if we didn't release it at the top.
    //
    if (!(devExtension->state & EARLY_RELEASE)) {
    
        IoReleaseCancelSpinLock(cancelIrql);
    }
    //
    // now if we found an irp to cancel, cancek it
    //
    if (cancelIrp) {

        //
        // this is ours!
        //
        if (Irp != cancelIrp) {

         KdPrint(("\t\t\tCanceled Irp %x (called Irp %x) queue position %d\n", 
              cancelIrp, Irp, position));

        } else {

         KdPrint(("\t\t\tCanceled Irp %x queue position %d\n", 
              cancelIrp, position));

        }
        
        cancelIrp->IoStatus.Status = STATUS_CANCELLED;
        cancelIrp->IoStatus.Information = 0;
        IoCompleteRequest(cancelIrp, IO_NO_INCREMENT);

    } else {

        KdPrint(("\t\t\tCancel(Irp %x) no cancelable irp found\n", Irp));
    }

}